/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#ifndef FLOAT33_H
#define FLOAT33_H

#include "float3.h"
#include <math.h>

/* Generate common floatXY funcs */
#define FLOATX_DIMENSION__ 3
#define FLOATY_DIMENSION__ 3
#include "floatXY.h"

/* Specific float33 funcs */
static FINLINE float*
f33
  (float* dst,
   const float a, const float b, const float c,
   const float d, const float e, const float f,
   const float g, const float h, const float i)
{
  ASSERT(dst);
  dst[0] = a; dst[1] = b; dst[2] = c;
  dst[3] = d; dst[4] = e; dst[5] = f;
  dst[6] = g; dst[7] = h; dst[8] = i;
  return dst;
}

static FINLINE float
f33_det(const float* mat)
{
  float tmp[3];
  f3_cross(tmp, f33_col_cptr(mat, 0), f33_col_cptr(mat, 1));
  return f3_dot(f33_col_cptr(mat, 2), tmp);
}

static FINLINE float
f33_invtrans(float* dst, const float* src)
{
  float f33[9];
  float det;
  f3_cross(f33_col_ptr(f33, 0), f33_col_cptr(src, 1), f33_col_cptr(src, 2));
  f3_cross(f33_col_ptr(f33, 1), f33_col_cptr(src, 2), f33_col_cptr(src, 0));
  f3_cross(f33_col_ptr(f33, 2), f33_col_cptr(src, 0), f33_col_cptr(src, 1));

  det = f3_dot(f33_col_cptr(f33, 2), f33_col_cptr(src, 2));
  f33_mul(dst, f33, 1.f / det);
  return det;
}

static FINLINE float
f33_inverse(float* dst, const float* src)
{
  float f33[9];
  const float det = f33_invtrans(f33, src);
  f33_transpose(dst, f33);
  return det;
}

static INLINE float*
f33_rotation /* XYZ norm */
  (float* dst,
   /* In radian */
   const float pitch,
   const float yaw,
   const float roll)
{
  const float c1 = (float)cos((double)pitch);
  const float c2 = (float)cos((double)yaw);
  const float c3 = (float)cos((double)roll);
  const float s1 = (float)sin((double)pitch);
  const float s2 = (float)sin((double)yaw);
  const float s3 = (float)sin((double)roll);
  ASSERT(dst);
  dst[0] = c2*c3; dst[1] = c1*s3 + c3*s1*s2; dst[2] = s1*s3 - c1*c3*s2;
  dst[3] =-c2*s3; dst[4] = c1*c3 - s1*s2*s3; dst[5] = c1*s2*s3 + c3*s1;
  dst[6] = s2;    dst[7] =-c2*s1;            dst[8] = c1*c2;
  return dst;
}

static INLINE float*
f33_rotation_axis_angle
  (float* dst,
   const float axis[3], /* Should be normalized */
   const float angle) /* In radian */
{
  const float c = (float)cos((double)angle);
  const float s = (float)sin((double)angle);
  const float C = 1 - c;
  ASSERT(dst && axis && f3_is_normalized(axis));

  dst[0] = axis[0] * axis[0] * C + c;
  dst[1] = axis[0] * axis[1] * C + s * axis[2];
  dst[2] = axis[0] * axis[2] * C - s * axis[1];

  dst[3] = axis[1] * axis[0] * C - s * axis[2];
  dst[4] = axis[1] * axis[1] * C + c;
  dst[5] = axis[1] * axis[2] * C + s * axis[0];

  dst[6] = axis[2] * axis[0] * C + s * axis[1];
  dst[7] = axis[2] * axis[1] * C - s * axis[0];
  dst[8] = axis[2] * axis[2] * C + c;
  return dst;
}

static INLINE float*
f33_rotation_pitch(float dst[9], const float pitch/* in radian */)
{
  const float c = (float)cos((double)pitch);
  const float s = (float)sin((double)pitch);
  ASSERT(dst);
  dst[0] = 1.f; dst[1] = 0.f; dst[2] = 0.f;
  dst[3] = 0.f; dst[4] = c;   dst[5] = s;
  dst[6] = 0.f; dst[7] =-s;   dst[8] = c;
  return dst;
}

static INLINE float*
f33_rotation_yaw(float dst[9], const float yaw/* in radian */)
{
  const float c = (float)cos((double)yaw);
  const float s = (float)sin((double)yaw);
  ASSERT(dst);
  dst[0] = c;   dst[1] = 0.f; dst[2] =-s;
  dst[3] = 0.f; dst[4] = 1.f; dst[5] = 0.f;
  dst[6] = s;   dst[7] = 0.f; dst[8] = c;
  return dst;
}

static INLINE float*
f33_rotation_roll(float dst[9], const float roll/* in radian */)
{
  const float c = (float)cos((double)roll);
  const float s = (float)sin((double)roll);
  ASSERT(dst);
  dst[0] = c;   dst[1] = s;   dst[2] = 0.f;
  dst[3] =-s;   dst[4] = c;   dst[5] = 0.f;
  dst[6] = 0.f; dst[7] = 0.f; dst[8] = 1.f;
  return dst;
}

static INLINE float*
f33_basis(float dst[9], const float N[3])
{
  float a[3], b[3], x[3], y[3], normal[3], len;
  ASSERT(N);
  f3_set(normal, N);
  f3_cross(a, f3(a, 1.f, 0.f, 0.f), normal);
  f3_cross(b, f3(b, 0.f, 1.f, 0.f), normal);
  len = f3_normalize(x, f3_dot(a, a) > f3_dot(b, b) ? a : b);
  if(len <= 0.f) {
    ASSERT(0 && "Degenerated normal");
    return f33_splat(dst, 0.f);
  }
  len = f3_normalize(y, f3_cross(y, normal, x));
  ASSERT(len > 0);
  f3_set(dst + 0, x);
  f3_set(dst + 3, y);
  f3_set(dst + 6, normal);
  return dst;
}

#endif /* FLOAT33_H */

