/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d.h"
#include "s3d_device_c.h"
#include "s3d_scene_c.h"
#include "s3d_shape_c.h"

#include <rsys/float3.h>
#include <rsys/float33.h>
#include <rsys/mem_allocator.h>
#include <rsys/mutex.h>

#include <algorithm>

/* Flag used to define session of enabled on instantiated scene */
#define S3D_INSTANCE (BIT(sizeof(int)*8 - 1))

struct ray_extended : public RTCRay {
  struct s3d_scene* scene;
  void* data; /* User defined data */
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static INLINE void
hit_setup(struct s3d_scene* scn, const RTCRay* ray, struct s3d_hit* hit)
{
  float w;
  char flip_surface = 0;

  ASSERT(scn && hit && ray);

  if((unsigned)ray->geomID == RTC_INVALID_GEOMETRY_ID) { /* No hit */
    *hit = S3D_HIT_NULL;
    return;
  }

  f3_set(hit->normal, ray->Ng);
  hit->distance = ray->tfar;

  hit->uv[0] = ray->u;
  hit->uv[1] = ray->v;
  w = 1.f - hit->uv[0] - hit->uv[1];
  ASSERT(w <= 1.f); /* This may not occurs */
  if(w < 0.f) { /* Handle precision error */
    if(hit->uv[0] > hit->uv[1]) hit->uv[0] += w;
    else hit->uv[1] += w;
    w = 0.f;
  }

  /* Embree stores on the u and v ray parameters the barycentric coordinates of
   * the hit with respect to the second and third triangle vertices,
   * respectively. The following code computes the barycentric coordinates of
   * the hit for the first and second triangle vertices */
  hit->uv[1] = hit->uv[0];
  hit->uv[0] = w;

  if((unsigned)ray->instID == RTC_INVALID_GEOMETRY_ID) {
    struct geometry* geom_mesh;
    ASSERT((unsigned)ray->geomID < darray_geom_size_get(&scn->embree2geoms));
    geom_mesh = darray_geom_data_get(&scn->embree2geoms)[ray->geomID];
    hit->prim.mesh__ = geom_mesh;
    hit->prim.inst__ = NULL;
    hit->prim.prim_id = ray->primID;
    hit->prim.geom_id = geom_mesh->name;
    hit->prim.inst_id = S3D_INVALID_ID;
    hit->prim.scene_prim_id = /* Compute the "scene space" primitive id */
      hit->prim.prim_id /* Mesh space */
    + geom_mesh->scene_prim_id_offset; /* Scene space */

  } else { /* The hit shape is instantiated */
    /* Retrieve the hit instance */
    struct geometry* geom_inst;
    struct geometry* geom_mesh;
    ASSERT((unsigned)ray->instID < darray_geom_size_get(&scn->embree2geoms));
    geom_inst = darray_geom_data_get(&scn->embree2geoms)[ray->instID];
    geom_mesh = scene_get_mesh(geom_inst->data.instance->scene, ray->geomID);
    hit->prim.mesh__ = geom_mesh;
    hit->prim.inst__ = geom_inst;
    hit->prim.prim_id = ray->primID;
    hit->prim.geom_id = geom_mesh->name;
    hit->prim.inst_id = geom_inst->name;
    hit->prim.scene_prim_id = /* Compute the "scene space" */
      hit->prim.prim_id /* Mesh space */
      + geom_mesh->scene_prim_id_offset /* Inst space */
      + geom_inst->scene_prim_id_offset; /* Scene space */

    flip_surface = geom_inst->flip_surface;
    ASSERT(hit->prim.inst__);
    ASSERT(((struct geometry*)hit->prim.inst__)->type == GEOM_INSTANCE);
  }
  ASSERT(hit->prim.mesh__);
  ASSERT(((struct geometry*)hit->prim.mesh__)->type == GEOM_MESH);

  /* Flip geometric normal with respect to the flip surface flag */
  flip_surface ^= ((struct geometry*)hit->prim.mesh__)->flip_surface;
  if(flip_surface) f3_minus(hit->normal, hit->normal);
}

/* Wrapper between an Embree and a Star-3D filter function */
static void
filter_wrapper(void* user_ptr, RTCRay& ray)
{
  struct s3d_hit hit;
  struct hit_filter* filter = (struct hit_filter*)user_ptr;
  struct ray_extended* ray_ex = static_cast<struct ray_extended*>(&ray);

  hit_setup(ray_ex->scene, &ray, &hit);
  if(filter->func(&hit, ray_ex->org, ray_ex->dir, ray_ex->data, filter->data)) {
    /* Discard the intersection */
    ray.geomID = RTC_INVALID_GEOMETRY_ID;
  }
}

static res_T
scene_sync
  (struct s3d_scene* scn,
   const int mask);/* combination of s3d_session_flag & S3D_INSTANCE */

static INLINE void
scene_geometry_flush_enable_state
  (struct s3d_scene* scn,
   struct geometry* geom, /* Cached geometry */
   const struct s3d_shape* shape) /* New  shape */
{
  ASSERT(scn && geom && shape);
  if(geom->is_enabled == shape->is_enabled)
    return;

  geom->is_enabled = shape->is_enabled;
  if(geom->is_enabled) {
    rtcEnable(scn->rtc_scn, geom->irtc);
  } else {
    rtcDisable(scn->rtc_scn, geom->irtc);
  }
  scn->is_rtc_scn_outdated = 1;
}

static res_T
scene_register_embree_geometry(struct s3d_scene* scn, struct geometry* geom)
{
  ASSERT(scn && geom && (geom->type==GEOM_MESH || geom->type==GEOM_INSTANCE));

  /* Create the Embree geometry if it is not valid */
  if(geom->irtc == RTC_INVALID_GEOMETRY_ID) {
    switch(geom->type) {
      case GEOM_MESH:
        geom->irtc = rtcNewTriangleMesh(scn->rtc_scn, RTC_GEOMETRY_DYNAMIC,
          mesh_get_ntris(geom->data.mesh), mesh_get_nverts(geom->data.mesh));
        break;
      case GEOM_INSTANCE:
        geom->irtc = rtcNewInstance
          (scn->rtc_scn, geom->data.instance->scene->rtc_scn);
        break;
      default: FATAL("Unreachable code\n"); break;
    }
    if(geom->irtc == RTC_INVALID_GEOMETRY_ID)
      return RES_UNKNOWN_ERR;

    scn->is_rtc_scn_outdated = 1;
  }

  if(geom->irtc >= darray_geom_size_get(&scn->embree2geoms)) {
    res_T res = darray_geom_resize(&scn->embree2geoms, geom->irtc + 1);
    if(res != RES_OK) {
      rtcDeleteGeometry(scn->rtc_scn, geom->irtc);
      geom->irtc = RTC_INVALID_GEOMETRY_ID;
      return res;
    }
  }
  geometry_ref_get(geom);
  darray_geom_data_get(&scn->embree2geoms)[geom->irtc] = geom;
  return RES_OK;
}

static void
scene_session_clear(struct s3d_scene* scn)
{
  struct geometry** geoms;
  size_t ngeoms;
  size_t i;
  ASSERT(scn);

  ngeoms = darray_geom_size_get(&scn->embree2geoms);
  geoms = darray_geom_data_get(&scn->embree2geoms);
  FOR_EACH(i, 0, ngeoms) {
    if(!geoms[i]) continue;

    if(geoms[i]->type == GEOM_INSTANCE)
      scene_session_clear(geoms[i]->data.instance->scene);

    geometry_ref_put(geoms[i]);
  }
  darray_geom_clear(&scn->embree2geoms);
  scn->session_mask = 0;
}

static res_T
scene_register_mesh
  (struct s3d_scene* scn,
   struct s3d_shape* shape)
{
  struct geometry** pgeom = NULL;
  struct geometry* geom = NULL;
  size_t iattr;
  char upd_pos, upd_ids;

  res_T res = RES_OK;
  ASSERT(shape && shape->type == GEOM_MESH);

  /* Retrieve the cached geometry */
  pgeom = htable_geom_find(&scn->cached_geoms, &shape);
  if(pgeom) {
    geom = *pgeom;
  } else {
    res = geometry_create(scn->dev, &geom);
    if(res != RES_OK) goto error;
    res = mesh_create(scn->dev, &geom->data.mesh);
    if(res != RES_OK) goto error;
    geom->type = GEOM_MESH;
    res = htable_geom_set(&scn->cached_geoms, &shape, &geom);
    if(res != RES_OK) goto error;
    geom->name = shape->id.index;
  }

  /* Discard the geometry that is not geometrically valid */
  if(!shape->data.mesh->indices || !shape->data.mesh->attribs[S3D_POSITION]) {
    if(geom->irtc != RTC_INVALID_GEOMETRY_ID) {
      rtcDeleteGeometry(scn->rtc_scn, geom->irtc);
      geom->irtc = RTC_INVALID_GEOMETRY_ID;
      scn->is_rtc_scn_outdated = 1;
    }
    mesh_clear(geom->data.mesh);
    goto exit;
  }

  /* Define which geometry buffers were updated */
  upd_ids = geom->data.mesh->indices != shape->data.mesh->indices
    || ((shape->data.mesh->update_mask & INDEX_BUFFER) != 0);
  upd_pos = geom->data.mesh->attribs[S3D_POSITION] != shape->data.mesh->attribs[S3D_POSITION]
    || ((shape->data.mesh->update_mask & VERTEX_BUFFER) != 0);

  /* Get a reference onto the shape mesh indices */
  if(geom->data.mesh->indices != shape->data.mesh->indices) {
    if(geom->data.mesh->indices) {
      index_buffer_ref_put(geom->data.mesh->indices);
      geom->data.mesh->indices = NULL;
    }
    ASSERT(shape->data.mesh->indices);
    index_buffer_ref_get(shape->data.mesh->indices);
    geom->data.mesh->indices = shape->data.mesh->indices;
  }

  /* Get the reference onto the shape mesh attribs */
  FOR_EACH(iattr, 0, S3D_ATTRIBS_COUNT__) {
    if(geom->data.mesh->attribs[iattr] == shape->data.mesh->attribs[iattr])
      continue;

    if(geom->data.mesh->attribs[iattr]) {
      vertex_buffer_ref_put(geom->data.mesh->attribs[iattr]);
      geom->data.mesh->attribs[iattr] = NULL;
    }
    if(!shape->data.mesh->attribs[iattr])
      continue;

    vertex_buffer_ref_get(shape->data.mesh->attribs[iattr]);
    geom->data.mesh->attribs[iattr] = shape->data.mesh->attribs[iattr];
    geom->data.mesh->attribs_type[iattr] = shape->data.mesh->attribs_type[iattr];
  }

  /* The shape mesh was resize => the Embree geometry is no more valid */
  if(shape->data.mesh->resize_mask && geom->irtc != RTC_INVALID_GEOMETRY_ID) {
    rtcDeleteGeometry(scn->rtc_scn, geom->irtc);
    geom->irtc = RTC_INVALID_GEOMETRY_ID;
    scn->is_rtc_scn_outdated = 1;
  }

  /* Update the cached mesh states */
  geom->flip_surface = shape->flip_surface;
  geom->data.mesh->filter = shape->data.mesh->filter;

  res = scene_register_embree_geometry(scn, geom);
  if(res != RES_OK) goto error;

  if(upd_pos) { /* Update the Embree vertex buffer if necessary */
    rtcSetBuffer(scn->rtc_scn, geom->irtc, RTC_VERTEX_BUFFER,
      mesh_get_pos(geom->data.mesh), 0, sizeof(float[3]));
    rtcUpdateBuffer(scn->rtc_scn, geom->irtc, RTC_VERTEX_BUFFER);
    scn->is_rtc_scn_outdated = 1;
  }
  if(upd_ids) { /* Update the Embree index buffer if necessary */
    rtcSetBuffer(scn->rtc_scn, geom->irtc, RTC_INDEX_BUFFER,
      mesh_get_ids(geom->data.mesh), 0, sizeof(uint32_t[3]));
    rtcUpdateBuffer(scn->rtc_scn, geom->irtc, RTC_INDEX_BUFFER);
    scn->is_rtc_scn_outdated = 1;
  }
  if(geom->data.mesh->filter.func) { /* Upd the filter func of the mesh */
    rtcSetIntersectionFilterFunction(scn->rtc_scn, geom->irtc, filter_wrapper);
    rtcSetUserData(scn->rtc_scn, geom->irtc, &geom->data.mesh->filter);
  }

  scene_geometry_flush_enable_state(scn, geom, shape);

  /* Flush the shape mesh states */
  shape->data.mesh->resize_mask = 0;
  shape->data.mesh->update_mask = 0;

exit:
  return res;
error:
  goto exit;
}

static res_T
scene_register_instance
  (struct s3d_scene* scn,
   struct s3d_shape* shape,
   const int session_mask)
{
  struct geometry** pgeom = NULL;
  struct geometry* geom = NULL;
  res_T res = RES_OK;
  ASSERT(scn && shape && shape->type == GEOM_INSTANCE);

  /* Recursuvely update the scene */
  res = scene_sync(shape->data.instance->scene,
    session_mask|S3D_INSTANCE);
  if(res != RES_OK) goto error;

  pgeom = htable_geom_find(&scn->cached_geoms, &shape);
  /* Create the scene instance of the geometry if necessary */
  if(pgeom) {
    geom = *pgeom;
  } else {
    res = geometry_create(scn->dev, &geom);
    if(res != RES_OK) goto error;
    geom->type = GEOM_INSTANCE;
    res = instance_create(shape->data.instance->scene, &geom->data.instance);
    if(res != RES_OK) goto error;
    res = htable_geom_set(&scn->cached_geoms, &shape, &geom);
    if(res != RES_OK) goto error;
    geom->name = shape->id.index;
  }
  /* Update the cached instance states */
  ASSERT(geom->data.instance->scene == shape->data.instance->scene);
  f33_set(geom->data.instance->transform, shape->data.instance->transform);
  f3_set(geom->data.instance->transform + 9, shape->data.instance->transform + 9);
  geom->flip_surface = shape->flip_surface;

  /* The instance cannot contain instances, i.e. one instancing level is
   * supported */
  if(geom->data.instance->scene->instances_count != 0) {
    res = RES_BAD_ARG;
    goto error;
  }

  /* Create the Embree instance */
  res = scene_register_embree_geometry(scn, geom);
  if(res != RES_OK) goto error;

  /* Update the Embree instance transformation */
  if(shape->data.instance->update_transform) {
    rtcSetTransform
      (scn->rtc_scn,
       geom->irtc,
       RTC_MATRIX_COLUMN_MAJOR,
       geom->data.instance->transform);
    scn->is_rtc_scn_outdated = 1;
  }

  scene_geometry_flush_enable_state(scn, geom, shape);
  shape->data.instance->update_transform = 0; /* Flush instance state */

exit:
  return res;
error:
  goto exit;
}

static void
scene_detach_shape(struct s3d_scene* scn, struct s3d_shape* shape)
{
  struct geometry** pgeom;
  ASSERT(scn && shape && !is_list_empty(&shape->scene_attachment));
  ASSERT(shape->type == GEOM_MESH || shape->type == GEOM_INSTANCE);
  ASSERT(scn->session_mask == 0);

  pgeom = htable_geom_find(&scn->cached_geoms, &shape);
  if(pgeom) { /* Remove the cached shape mesh */
    struct geometry* geom = *pgeom;
    if(geom->irtc != RTC_INVALID_GEOMETRY_ID) {
      rtcDeleteGeometry(scn->rtc_scn, geom->irtc);
      geom->irtc = RTC_INVALID_GEOMETRY_ID;
      scn->is_rtc_scn_outdated = 1;
    }
    geometry_ref_put(geom);
    htable_geom_erase(&scn->cached_geoms, &shape);
  }
  list_del(&shape->scene_attachment);

  if(shape->type == GEOM_INSTANCE) {
    ASSERT(scn->instances_count != 0);
    --scn->instances_count;
  }

  S3D(shape_ref_put(shape));
}

static res_T
scene_compute_cdf(struct s3d_scene* scn)
{
  struct list_node* node;
  struct s3d_shape* shape;
  struct geometry** pgeom;
  struct geometry* geom;
  size_t len;
  float area = 0.f;
  res_T res = RES_OK;
  ASSERT(scn);

  darray_fltui_clear(&scn->cdf);

  LIST_FOR_EACH(node, &scn->shapes) {
    shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
    pgeom = htable_geom_find(&scn->cached_geoms, &shape);
    ASSERT(pgeom != NULL);
    geom = *pgeom;
    struct fltui fltui;

    if(!geom->is_enabled) continue;

    switch(geom->type) {
      case GEOM_MESH:
        res = mesh_compute_cdf(geom->data.mesh);
        if(res != RES_OK) goto error;
        len = darray_float_size_get(&geom->data.mesh->cdf);
        if(len) {
          area += darray_float_cdata_get(&geom->data.mesh->cdf)[len - 1];
        }
        break;
      case GEOM_INSTANCE:
        res = scene_compute_cdf(geom->data.instance->scene);
        if(res != RES_OK) goto error;
        len = darray_fltui_size_get(&geom->data.instance->scene->cdf);
        if(len) {
          area += darray_fltui_cdata_get
            (&geom->data.instance->scene->cdf)[len - 1].flt;
        }
        break;
      default: FATAL("Unreachable code\n"); break;
    }
    fltui.ui = geom->irtc;
    fltui.flt = area;
    if(len) {
      res = darray_fltui_push_back(&scn->cdf, &fltui);
      if(res != RES_OK) goto error;
    }
  }
exit:
  return res;
error:
  darray_fltui_clear(&scn->cdf);
  goto exit;
}

static void
scene_compute_aabb(struct s3d_scene* scn)
{
  struct list_node* node;
  struct s3d_shape* shape;
  struct geometry** pgeom;
  struct geometry* geom;
  struct instance* inst;
  float lower[3], upper[3];

  f3_splat(scn->lower, FLT_MAX);
  f3_splat(scn->upper,-FLT_MAX);

  LIST_FOR_EACH(node, &scn->shapes) {
    shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
    pgeom = htable_geom_find(&scn->cached_geoms, &shape);
    ASSERT(pgeom != NULL);
    geom = *pgeom;

    if(!geom->is_enabled) continue;

    switch(geom->type) {
      case GEOM_MESH: mesh_compute_aabb(geom->data.mesh, lower, upper); break;
      case GEOM_INSTANCE:
        inst = geom->data.instance;
        scene_compute_aabb(inst->scene);
        /* Transform local scene AABB in world space */
        f33_mulf3(lower, inst->transform, inst->scene->lower);
        f33_mulf3(upper, inst->transform, inst->scene->upper);
        f3_add(lower, inst->transform + 9, lower);
        f3_add(upper, inst->transform + 9, upper);
        break;
      default: FATAL("Unreachable code\n"); break;
    }
    f3_min(scn->lower, scn->lower, lower);
    f3_max(scn->upper, scn->upper, upper);
  }
}

static FINLINE bool
operator < (const struct fltui& it, const float val)
{
  /* This operator is used by the std::lower_bound algorithm that returns an
   * iterator to the first element that is not less than val while one expect
   * an iterator on the first element that is not less *or equal* than val.
   * That's why we use <= rather than < */
  return it.flt <= val;
}

static res_T
scene_compute_nprims_cdf
  (struct s3d_scene* scn,
   const char store_cdf)
{
  struct list_node* node;
  struct s3d_shape* shape;
  struct geometry** pgeom;
  struct geometry* geom;
  size_t len;
  unsigned nprims;
  res_T res = RES_OK;
  ASSERT(scn);

  darray_nprims_cdf_clear(&scn->nprims_cdf);

  nprims = 0;
  LIST_FOR_EACH(node, &scn->shapes) {
    shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
    pgeom = htable_geom_find(&scn->cached_geoms, &shape);
    ASSERT(pgeom != NULL);
    geom = *pgeom;
    struct nprims_cdf cdf;

    if(!geom->is_enabled) continue;

    geom->scene_prim_id_offset = nprims;
    switch(geom->type) {
      case GEOM_MESH:
        len = mesh_get_ntris(geom->data.mesh);
        nprims += (unsigned)len;
        break;
      case GEOM_INSTANCE:
        res = scene_compute_nprims_cdf(geom->data.instance->scene, store_cdf);
        if(res != RES_OK) goto error;
        len = darray_nprims_cdf_size_get(&geom->data.instance->scene->nprims_cdf);
        if(len) {
          nprims += darray_nprims_cdf_cdata_get
            (&geom->data.instance->scene->nprims_cdf)[len - 1].nprims;
        }
        break;
      default: FATAL("Unreachable code\n"); break;
    }

    cdf.nprims = nprims;
    cdf.irtc = geom->irtc;
    if(store_cdf && len) {
      res = darray_nprims_cdf_push_back(&scn->nprims_cdf, &cdf);
      if(res != RES_OK) goto error;
    }
  }
exit:
  return res;
error:
  darray_nprims_cdf_clear(&scn->nprims_cdf);
  goto exit;
}

static FINLINE bool
operator < (const struct nprims_cdf& it, const size_t iprim)
{
  /* This operator is used by the std::lower_bound algorithm that returns an
   * iterator to the first element that is not less than iprim while one expect
   * an iterator on the first element that is not less *or equal* than iprim.
   * That's why we use <= rather than < */
  return it.nprims <= iprim;
}

static res_T
scene_sync(struct s3d_scene* scn, const int session_mask)
{
  struct list_node* node;
  struct s3d_shape* shape;
  res_T res = RES_OK;
  ASSERT(scn);

  if((session_mask & S3D_INSTANCE) != 0
  && (scn->session_mask & S3D_INSTANCE) != 0) {
    /* The scene was already synced as an instance. Discard sync process */
    return RES_OK;
  } else if(scn->session_mask != 0) {
    /* The scene cannot be synced several times exepted if it is instantiated */
    res = RES_BAD_OP;
    goto error;
  }

  if((session_mask & S3D_INSTANCE) == 0 && scn->session_mask != 0) {
    res = RES_BAD_OP;
    goto error;
  }

  LIST_FOR_EACH(node, &scn->shapes) {
    shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
    switch(shape->type) {
      case GEOM_INSTANCE:
        /* One instancing level is supported */
        if((session_mask & S3D_INSTANCE) != 0) {
          res = RES_BAD_ARG;
          goto error;
        }
        res = scene_register_instance(scn, shape, session_mask);
        break;
      case GEOM_MESH:
        res = scene_register_mesh(scn, shape);
        break;
      default: FATAL("Unreachable code\n"); break;
    }
    if(res != RES_OK)
      goto error;
  }
  if((session_mask & S3D_SAMPLE) != 0) {
    res = scene_compute_cdf(scn);
    if(res != RES_OK) goto error;
  }
  if((session_mask & S3D_GET_PRIMITIVE) != 0) {
    res = scene_compute_nprims_cdf(scn, 1);
  } else {
    res = scene_compute_nprims_cdf(scn, 0);
  }
  if(res != RES_OK) goto error;

  if((session_mask & S3D_TRACE) != 0 && scn->is_rtc_scn_outdated) {
    rtcCommit(scn->rtc_scn);
    scn->is_rtc_scn_outdated = 0;
  }
  scn->session_mask = session_mask;

  scene_compute_aabb(scn);

exit:
  return res;
error:
  goto exit;
}

static void
scene_release(ref_T* ref)
{
  struct s3d_scene* scn;
  struct s3d_device* dev;
  ASSERT(ref);
  scn = CONTAINER_OF(ref, struct s3d_scene, ref);
  S3D(scene_clear(scn));
  dev = scn->dev;
  scene_session_clear(scn);
  if(scn->rtc_scn) rtcDeleteScene(scn->rtc_scn);
  htable_geom_release(&scn->cached_geoms);
  darray_geom_release(&scn->embree2geoms);
  darray_fltui_release(&scn->cdf);
  darray_nprims_cdf_release(&scn->nprims_cdf);
  MEM_RM(dev->allocator, scn);
  S3D(device_ref_put(dev));
}

/*******************************************************************************
 * Exported s3d_scene functions
 ******************************************************************************/
res_T
s3d_scene_create(struct s3d_device* dev, struct s3d_scene** out_scn)
{
  struct s3d_scene* scn = NULL;
  res_T res = RES_OK;

  if(!dev || !out_scn) {
    res = RES_BAD_ARG;
    goto error;
  }
  scn = (struct s3d_scene*)MEM_CALLOC
    (dev->allocator, 1, sizeof(struct s3d_scene));
  if(!scn) {
    res = RES_MEM_ERR;
    goto error;
  }
  list_init(&scn->shapes);
  htable_geom_init(dev->allocator, &scn->cached_geoms);
  darray_geom_init(dev->allocator, &scn->embree2geoms);
  darray_fltui_init(dev->allocator, &scn->cdf);
  darray_nprims_cdf_init(dev->allocator, &scn->nprims_cdf);
  ref_init(&scn->ref);
  S3D(device_ref_get(dev));
  scn->dev = dev;
  scn->session_mask = 0;
  scn->rtc_scn = rtcDeviceNewScene
    (dev->rtc, RTC_SCENE_DYNAMIC | RTC_SCENE_INCOHERENT | RTC_SCENE_ROBUST,
     RTC_INTERSECT1 | RTC_INTERSECT4);
  if(!scn->rtc_scn) {
    res = RES_MEM_ERR;
    goto error;
  }
  f3_splat(scn->lower, FLT_MAX);
  f3_splat(scn->upper,-FLT_MAX);

exit:
  if(out_scn) *out_scn = scn;
  return res;
error:
  if(scn) {
    S3D(scene_ref_put(scn));
    scn = NULL;
  }
  goto exit;
}

res_T
s3d_scene_ref_get(struct s3d_scene* scn)
{
  if(!scn) return RES_BAD_ARG;
  ref_get(&scn->ref);
  return RES_OK;
}

res_T
s3d_scene_ref_put(struct s3d_scene* scn)
{
  if(!scn) return RES_BAD_ARG;
  ref_put(&scn->ref, scene_release);
  return RES_OK;
}

res_T
s3d_scene_instantiate(struct s3d_scene* scn, struct s3d_shape** out_shape)
{
  struct s3d_shape* shape = NULL;
  res_T res = RES_OK;

  if(!scn || !out_shape) {
    res = RES_BAD_ARG;
    goto error;
  }

  res = shape_create(scn->dev, &shape);
  if(res != RES_OK)
    goto error;

  shape->type = GEOM_INSTANCE;
  res = instance_create(scn, &shape->data.instance);
  if(res != RES_OK)
    goto error;

exit:
  if(out_shape)
    *out_shape = shape;
  return res;
error:
  if(shape) {
    S3D(shape_ref_put(shape));
    shape = NULL;
  }
  goto exit;
}

res_T
s3d_scene_attach_shape(struct s3d_scene* scn, struct s3d_shape* shape)
{
  if(!scn || !shape)
    return RES_BAD_ARG;
  if(!is_list_empty(&shape->scene_attachment))
    return RES_BAD_ARG;
  if(shape->type == GEOM_INSTANCE && shape->data.instance->scene == scn)
    return RES_BAD_ARG;

  list_add_tail(&scn->shapes, &shape->scene_attachment);
  S3D(shape_ref_get(shape));
  scn->instances_count += shape->type == GEOM_INSTANCE;
  return RES_OK;
}

res_T
s3d_scene_detach_shape(struct s3d_scene* scn, struct s3d_shape* shape)
{
  char is_attached;
  if(!scn || !shape) return RES_BAD_ARG;
  if(scn->session_mask != 0) return RES_BAD_OP;
  if(!(S3D(shape_is_attached(shape, &is_attached)), is_attached))
    return RES_BAD_ARG;
#ifndef NDEBUG
  { /* Check the the shape is attached to `scn' */
    struct list_node* node;
    char is_found = 0;
    LIST_FOR_EACH(node, &scn->shapes) {
      if(node == &shape->scene_attachment)
        is_found = 1;
    }
    ASSERT(is_found);
  }
#endif
  scene_detach_shape(scn, shape);
  return RES_OK;
}

res_T
s3d_scene_clear(struct s3d_scene* scn)
{
  struct list_node* node, *tmp;
  if(!scn) return RES_BAD_ARG;
  if(scn->session_mask != 0) return RES_BAD_OP;
  LIST_FOR_EACH_SAFE(node, tmp, &scn->shapes) {
    struct s3d_shape* shape = CONTAINER_OF
      (node, struct s3d_shape, scene_attachment);
    scene_detach_shape(scn, shape);
  }
  return RES_OK;
}

res_T
s3d_scene_begin_session(struct s3d_scene* scn, const int session_mask)
{
  if(!scn)
    return RES_BAD_ARG;
  if(!(session_mask&S3D_TRACE)
  && !(session_mask&S3D_SAMPLE)
  && !(session_mask&S3D_GET_PRIMITIVE))
    return RES_BAD_ARG;
  return scene_sync(scn, session_mask);
}

res_T
s3d_scene_end_session(struct s3d_scene* scn)
{
  if(!scn)
    return RES_BAD_ARG;
  if(scn->session_mask & S3D_INSTANCE || !scn->session_mask)
    return RES_BAD_OP;
  scene_session_clear(scn);
  return RES_OK;
}

res_T
s3d_scene_get_session_mask(struct s3d_scene* scn, int* session_mask)
{
  if(!scn || !session_mask)
    return RES_BAD_ARG;
  *session_mask = scn->session_mask & (~S3D_INSTANCE);
  return RES_OK;
}

res_T
s3d_scene_trace_ray
  (struct s3d_scene* scn,
   const float org[3],
   const float dir[3],
   const float range[2],
   void* ray_data,
   struct s3d_hit* hit)
{
  struct ray_extended ray_ex;
  if(!scn || !org || !dir || !range || !hit)
    return RES_BAD_ARG;
  if(!f3_is_normalized(dir))
    return RES_BAD_ARG;
  if((scn->session_mask & S3D_TRACE) == 0)
    return RES_BAD_OP;
  if(range[0] > range[1]) {
    *hit = S3D_HIT_NULL;
    return RES_OK;
  }

  f3_set(ray_ex.org, org);
  f3_set(ray_ex.dir, dir);
  ray_ex.tnear = range[0];
  ray_ex.tfar = range[1];
  ray_ex.geomID = RTC_INVALID_GEOMETRY_ID;
  ray_ex.primID = RTC_INVALID_GEOMETRY_ID;
  ray_ex.instID = RTC_INVALID_GEOMETRY_ID;
  ray_ex.mask = 0xFFFFFFFF;
  ray_ex.time = 0.f;
  ray_ex.scene = scn;
  ray_ex.data = ray_data;

  rtcIntersect(scn->rtc_scn, ray_ex);

  hit_setup(scn, &ray_ex, hit);

  return  RES_OK;
}

res_T
s3d_scene_trace_rays
  (struct s3d_scene* scn,
   const size_t nrays,
   const int mask,
   const float* origins,
   const float* directions,
   const float* ranges,
   void* rays_data,
   const size_t sizeof_ray_data,
   struct s3d_hit* hits)
{
  size_t iray;
  size_t iorg, idir, irange, idata;
  size_t org_step, dir_step, range_step, data_step;
  res_T res = RES_OK;

  if(!scn) return RES_BAD_ARG;
  if(!nrays) return RES_OK;

  org_step = mask & S3D_RAYS_SINGLE_ORIGIN ? 0 : 3;
  dir_step = mask & S3D_RAYS_SINGLE_DIRECTION ? 0 : 3;
  range_step = mask & S3D_RAYS_SINGLE_RANGE ? 0 : 2;
  data_step = (mask & S3D_RAYS_SINGLE_DATA) || !rays_data ? 0 : sizeof_ray_data;
  iorg = idir = irange = idata = 0;

  FOR_EACH(iray, 0, nrays) {
    res = s3d_scene_trace_ray(scn, origins+iorg, directions+idir,
      ranges+irange, (char*)rays_data+idata, hits+iray);
    if(UNLIKELY(res != RES_OK)) break;
    iorg += org_step;
    idir += dir_step;
    irange += range_step;
    idata += data_step;
  }
  return res;
}

res_T
s3d_scene_sample
  (struct s3d_scene* scn,
   const float u,
   const float v,
   const float w,
   struct s3d_primitive* primitive, /* sampled primitive */
   float st[2])
{
  struct geometry* geom;
  const struct fltui* fltui_begin, *fltui_end, *fltui_found;
  const float* flt_begin, *flt_end, *flt_found;
  size_t igeom;
  float f;
  res_T res = RES_OK;

  if(!scn || !primitive || !st) {
    res = RES_BAD_ARG;
    goto error;
  }
  /* Expecting canonic numbers */
  if(u < 0.f || u >= 1.f || v < 0.f || v >= 1.f || w < 0.f || w >= 1.f) {
    res = RES_BAD_ARG;
    goto error;
  }
  if((scn->session_mask & S3D_SAMPLE) == 0) {
    res = RES_BAD_OP;
    goto error;
  }

  /* Find the sampled geometry */
  if(darray_fltui_size_get(&scn->cdf) == 0) {
    /* No geometry to sample */
    *primitive = S3D_PRIMITIVE_NULL;
    goto exit;
  } else if(darray_fltui_size_get(&scn->cdf) == 1) {
    igeom = darray_fltui_cdata_get(&scn->cdf)[0].ui;
    f = u * darray_fltui_cdata_get(&scn->cdf)[0].flt;/* Map u to the CDF bounds */
  } else {
    fltui_begin = darray_fltui_cdata_get(&scn->cdf);
    fltui_end = fltui_begin + darray_fltui_size_get(&scn->cdf);
    f = u * fltui_end[-1].flt; /* Map u to the CDF bounds */
    fltui_found = std::lower_bound(fltui_begin, fltui_end, f);
    ASSERT(fltui_found != fltui_end);
    igeom = fltui_found->ui;

    /* Transform u to the geometry CDF bounds */
    if(fltui_found != fltui_begin)
      f -= fltui_found[-1].flt;
  }
  geom = darray_geom_data_get(&scn->embree2geoms)[igeom];
  ASSERT(geom);

  if(geom->type == GEOM_MESH) {
    primitive->inst__ = NULL;
    primitive->inst_id = S3D_INVALID_ID;
    primitive->scene_prim_id = 0;
  } else {
    /* Find the sampled instantiated geometry */
    ASSERT(geom->type == GEOM_INSTANCE);
    primitive->inst__ = geom;
    primitive->inst_id = geom->name;
    primitive->scene_prim_id = geom->scene_prim_id_offset;
    if(darray_fltui_size_get(&geom->data.instance->scene->cdf) == 1) {
      igeom = darray_fltui_cdata_get(&geom->data.instance->scene->cdf)[0].ui;
    } else {
      fltui_begin = darray_fltui_cdata_get(&geom->data.instance->scene->cdf);
      fltui_end = fltui_begin + darray_fltui_size_get(&geom->data.instance->scene->cdf);
      fltui_found = std::lower_bound(fltui_begin, fltui_end, f);
      ASSERT(fltui_found != fltui_end);
      igeom = fltui_found->ui;

      /* Transform u to the geometry CDF bounds */
      if(fltui_found != fltui_begin)
        f -= fltui_found[-1].flt;
    }
    geom = darray_geom_data_get
      (&geom->data.instance->scene->embree2geoms)[igeom];
    ASSERT(geom);
  }
  ASSERT(geom->type == GEOM_MESH);

  /* Find the sampled triangle */
  primitive->mesh__ = geom;
  primitive->geom_id = geom->name;
  primitive->scene_prim_id += geom->scene_prim_id_offset;
  flt_begin = darray_float_cdata_get(&geom->data.mesh->cdf);
  flt_end = flt_begin + darray_float_size_get(&geom->data.mesh->cdf);
  flt_found = std::lower_bound(flt_begin, flt_end, f);
  ASSERT(flt_found != flt_end);

  primitive->prim_id = (unsigned)(flt_found - flt_begin);
  primitive->scene_prim_id += primitive->prim_id;
  S3D(primitive_sample(primitive, v, w, st));

exit:
  return res;
error:
  goto exit;
}

res_T
s3d_scene_get_primitive
  (struct s3d_scene* scn, const unsigned iprim, struct s3d_primitive* prim)
{
  struct geometry* geom;
  const struct nprims_cdf* begin, *end, *found;
  size_t nprims;
  size_t igeom;
  size_t i;
  res_T res = RES_OK;

  if(!scn || !prim) {
    res = RES_BAD_ARG;
    goto error;
  }
  if((scn->session_mask & S3D_GET_PRIMITIVE) == 0) {
    res = RES_BAD_OP;
    goto error;
  }
  S3D(scene_primitives_count(scn, &nprims));
  if(iprim  >= nprims) {
    res = RES_BAD_ARG;
    goto error;
  }

  i = iprim;
  if(darray_nprims_cdf_size_get(&scn->nprims_cdf) == 1) {
    igeom = darray_nprims_cdf_cdata_get(&scn->nprims_cdf)[0].irtc;
  } else {
    begin = darray_nprims_cdf_cdata_get(&scn->nprims_cdf);
    end = begin + darray_nprims_cdf_size_get(&scn->nprims_cdf);
    found = std::lower_bound(begin, end, i);
    ASSERT(found != end);
    igeom = found->irtc;
    if(found != begin) {
      ASSERT(i >= found[-1].nprims);
      i -= found[-1].nprims;
    }
  }
  geom = darray_geom_data_get(&scn->embree2geoms)[igeom];
  ASSERT(geom);

  if(geom->type == GEOM_MESH) {
    prim->inst__ = NULL;
    prim->inst_id = S3D_INVALID_ID;
    prim->scene_prim_id = 0;
  } else {
    ASSERT(geom->type == GEOM_INSTANCE);
    prim->inst__ = geom;
    prim->inst_id = geom->name;
    prim->scene_prim_id = geom->scene_prim_id_offset;
    if(darray_nprims_cdf_size_get(&geom->data.instance->scene->nprims_cdf)==1) {
      igeom = darray_nprims_cdf_cdata_get
        (&geom->data.instance->scene->nprims_cdf)[0].irtc;
    } else {
      begin = darray_nprims_cdf_cdata_get
        (&geom->data.instance->scene->nprims_cdf);
      end = begin + darray_nprims_cdf_size_get
        (&geom->data.instance->scene->nprims_cdf);
      found = std::lower_bound(begin, end, i);
      ASSERT(found != end);
      igeom = found->irtc;
      if(found != begin) {
        ASSERT(i >= found[-1].nprims);
        i -= found[-1].nprims;
      }
    }
    geom = darray_geom_data_get
      (&geom->data.instance->scene->embree2geoms)[igeom];
    ASSERT(geom);
  }
  ASSERT(geom->type == GEOM_MESH);
  ASSERT(i < mesh_get_ntris(geom->data.mesh));
  prim->mesh__ = geom;
  prim->geom_id = geom->name;
  prim->prim_id = (unsigned)i;
  prim->scene_prim_id += geom->scene_prim_id_offset;
  prim->scene_prim_id += prim->prim_id;

exit:
  return res;
error:
  goto exit;
}

res_T
s3d_scene_primitives_count(struct s3d_scene* scn, size_t* prims_count)
{
  res_T res = RES_OK;

  if(!scn || !prims_count) {
    res = RES_BAD_ARG;
    goto error;
  }
  if(!scn->session_mask) {
    res = RES_BAD_OP;
    goto error;
  }
  if((scn->session_mask & S3D_GET_PRIMITIVE) != 0) {
    const size_t len = darray_nprims_cdf_size_get(&scn->nprims_cdf);
    if(!len) {
      *prims_count = 0;
    } else {
      *prims_count = darray_nprims_cdf_cdata_get(&scn->nprims_cdf)[len - 1].nprims;
    }
  } else {
    struct list_node* node;
    struct s3d_shape* shape;
    struct geometry** pgeom;
    struct geometry* geom;
    size_t inst_count;
    *prims_count = 0;
    LIST_FOR_EACH(node, &scn->shapes) {
      shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
      pgeom = htable_geom_find(&scn->cached_geoms, &shape);
      ASSERT(pgeom != NULL);
      geom = *pgeom;

      if(!geom->is_enabled) continue;

      switch(geom->type) {
        case GEOM_MESH:
          *prims_count += mesh_get_ntris(geom->data.mesh);
          break;
        case GEOM_INSTANCE:
          S3D(scene_primitives_count(geom->data.instance->scene, &inst_count));
          *prims_count += inst_count;
          break;
        default: FATAL("Unreachable code\n"); break;
      }
    }
  }
exit:
  return res;
error:
  goto exit;
}

res_T
s3d_scene_compute_area(struct s3d_scene* scn, float* area)
{
  res_T res = RES_OK;

  if(!scn || !area) {
    res = RES_BAD_ARG;
    goto error;
  }
  if(!scn->session_mask) {
    res = RES_BAD_OP;
    goto error;
  }

  if((scn->session_mask & S3D_SAMPLE) != 0) {
    /* Retrieve the overall scene area from the scene cumulative distribution
     * function. Note that the CDF stores the cumulateive triangle area
     * multiplied by 2; the real scene area is thus the CDF upper bound / 2 */
    size_t len = darray_fltui_size_get(&scn->cdf);
    if(!len) {
      *area = 0.f;
    } else {
      *area = darray_fltui_cdata_get(&scn->cdf)[len - 1].flt * 0.5f;
    }
  } else {
    struct list_node* node;
    struct s3d_shape* shape;
    struct geometry** pgeom;
    struct geometry* geom;
    float inst_area;

    *area = 0.f;
    LIST_FOR_EACH(node, &scn->shapes) {
      shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
      pgeom = htable_geom_find(&scn->cached_geoms, &shape);
      ASSERT(pgeom != NULL);
      geom = *pgeom;

      if(!geom->is_enabled) continue;

      switch(geom->type) {
        case GEOM_MESH:
          *area += mesh_compute_area(geom->data.mesh);
          break;
        case GEOM_INSTANCE:
          /* TODO take into account the instance scale factor */
          S3D(scene_compute_area(geom->data.instance->scene, &inst_area));
          *area += inst_area;
          break;
        default: FATAL("Unreachable code\n"); break;
      }
    }
  }

exit:
  return res;
error:
  goto exit;
}

res_T
s3d_scene_compute_volume(struct s3d_scene* scn, float* volume)
{
  struct list_node* node;
  struct s3d_shape* shape;
  struct geometry** pgeom;
  struct geometry* geom;
  res_T res = RES_OK;

  if(!scn || !volume) {
    res = RES_BAD_ARG;
    goto error;
  }
  if(!scn->session_mask) {
    res = RES_BAD_OP;
    goto error;
  }

  *volume = 0.f;
  LIST_FOR_EACH(node, &scn->shapes) {
    shape = CONTAINER_OF(node, struct s3d_shape, scene_attachment);
    pgeom = htable_geom_find(&scn->cached_geoms, &shape);
    ASSERT(pgeom != NULL);
    geom = *pgeom;

    if(!geom->is_enabled) continue;

    switch(geom->type) {
      case GEOM_MESH:
        *volume += mesh_compute_volume(geom->data.mesh, geom->flip_surface);
        break;
      case GEOM_INSTANCE:
        *volume += instance_compute_volume(geom->data.instance, geom->flip_surface);
        break;
      default: FATAL("Unreachable code\n"); break;
    }
  }

exit:
  return res;
error:
  goto exit;
}

res_T
s3d_scene_get_aabb(struct s3d_scene* scn, float lower[3], float upper[3])
{
  if(!scn || !lower || !upper) return RES_BAD_ARG;
  if(!scn->session_mask) return RES_BAD_OP;
  f3_set(lower, scn->lower);
  f3_set(upper, scn->upper);
  return RES_OK;
}

res_T
s3d_scene_get_device(struct s3d_scene* scn, struct s3d_device** dev)
{
  if(!scn || !dev) return RES_BAD_ARG;
  *dev = scn->dev;
  return RES_OK;
}

