/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d.h"
#include "test_s3d_cbox.h"
#include "test_s3d_utils.h"

#include <rsys/image.h>
#include <rsys/float3.h>

#include <string.h>

#define IMG_WIDTH 640
#define IMG_HEIGHT 480

struct camera {
  float pos[3];
  float x[3], y[3], z[3]; /* Basis */
};

static void
camera_init(struct camera* cam)
{
  const float pos[3] = { 178.f, -1000.f, 273.f };
  const float tgt[3] = { 178.f, 0.f, 273.f };
  const float up[3] = { 0.f, 0.f, 1.f };
  const float proj_ratio = (float)IMG_WIDTH/(float)IMG_HEIGHT;
  const float fov_x = (float)PI * 0.25f;
  float f = 0.f;
  ASSERT(cam);

  f3_set(cam->pos, pos);
  f = f3_normalize(cam->z, f3_sub(cam->z, tgt, pos)); NCHECK(f, 0);
  f = f3_normalize(cam->x, f3_cross(cam->x, cam->z, up)); NCHECK(f, 0);
  f = f3_normalize(cam->y, f3_cross(cam->y, cam->z, cam->x)); NCHECK(f, 0);
  f3_divf(cam->z, cam->z, (float)tan(fov_x*0.5f));
  f3_divf(cam->y, cam->y, proj_ratio);
}

static void
camera_ray
  (const struct camera* cam,
   const float pixel[2],
   float org[3],
   float dir[3])
{
  float x[3], y[3], f;
  ASSERT(cam && pixel && org && dir);

  f3_mulf(x, cam->x, pixel[0]*2.f - 1.f);
  f3_mulf(y, cam->y, pixel[1]*2.f - 1.f);
  f3_add(dir, f3_add(dir, x, y), cam->z);
  f = f3_normalize(dir, dir); NCHECK(f, 0);
  f3_set(org, cam->pos);
}

static int
filter_func
  (const struct s3d_hit* hit,
   const float pos[3],
   const float dir[3],
   void* ray_data,
   void* filter_data)
{
  NCHECK(hit, NULL);
  NCHECK(pos, NULL);
  NCHECK(dir, NULL);
  CHECK((uintptr_t)ray_data, 0xDEADBEEF);
  CHECK((uintptr_t)filter_data, 0xDECAFBAD);
  CHECK(S3D_HIT_NONE(hit), 0);
  return hit->prim.prim_id % 2 == 0;
}

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct s3d_device* dev;
  struct s3d_hit hit;
  struct s3d_scene* scn;
  struct s3d_scene* scn2;
  struct s3d_shape* inst;
  struct s3d_shape* walls;
  struct s3d_shape* walls_copy;
  struct s3d_shape* tall_block;
  struct s3d_shape* short_block;
  struct s3d_vertex_data attribs;
  struct s3d_primitive prims[30];
  struct camera cam;
  struct cbox_desc desc;
  unsigned char* img = NULL;
  unsigned ntris, nverts;
  size_t nprims;
  size_t ix, iy;
  float lower[3], upper[3];
  float org[3] = { 0.f, 0.f, 0.f };
  float dir[3] = { 0.f, 1.f, 0.f };
  float range[2] = { 0.f, FLT_MAX };
  unsigned inst_id;
  unsigned walls_id;
  unsigned tall_block_id;
  unsigned short_block_id;
  size_t i;
  char filter = 0;
  char* img_name = NULL;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  if(argc > 1) {
    if(!strcmp(argv[1], "filter")) {
      filter = 1;
    } else {
      img_name = argv[1];
    }
  }
  if(!img_name && argc > 2) {
    img_name = argv[2];
  }
  if(img_name) {
    img = MEM_ALLOC(&allocator, 3 * IMG_WIDTH * IMG_HEIGHT);
    NCHECK(img, NULL);
  }

  CHECK(s3d_device_create(NULL, &allocator, 1, &dev), RES_OK);
  CHECK(s3d_scene_create(dev, &scn), RES_OK);

  attribs.usage = S3D_POSITION;
  attribs.type = S3D_FLOAT3;
  attribs.get = cbox_get_position;

  ntris = cbox_walls_ntris;
  nverts = cbox_walls_nverts;
  desc.vertices = cbox_walls;
  desc.indices = cbox_walls_ids;
  CHECK(s3d_shape_create_mesh(dev, &walls), RES_OK);
  CHECK(s3d_mesh_setup_indexed_vertices
    (walls, ntris, cbox_get_ids, nverts, &attribs, 1, &desc), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, walls), RES_OK);
  CHECK(s3d_shape_ref_put(walls), RES_OK);

  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_trace_ray(NULL, NULL, NULL, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, NULL, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, NULL, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, NULL, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, dir, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, dir, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, dir, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, dir, NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, NULL, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, NULL, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, NULL, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, NULL, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, dir, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, dir, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, dir, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, dir, range, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, NULL, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, NULL, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, NULL, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, NULL, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, dir, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, dir, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, dir, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, dir, NULL, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, NULL, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, NULL, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, NULL, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, NULL, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, NULL, dir, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, NULL, dir, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(NULL, org, dir, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_trace_ray(scn, org, dir, range, NULL, &hit), RES_OK);
  CHECK(s3d_scene_trace_ray(scn, org, dir, range, NULL, &hit), RES_OK);
  f3(dir, 1.f, 1.f, 1.f);
  CHECK(s3d_scene_trace_ray(scn, org, dir, range, NULL, &hit), RES_BAD_ARG);
  CHECK(s3d_scene_end_session(scn), RES_OK);

  f3(dir, 0.f, 1.f, 0.f);
  CHECK(s3d_scene_trace_ray(scn, org, dir, range, NULL, &hit), RES_BAD_OP);
  CHECK(s3d_scene_clear(scn), RES_OK);

  /* Update the inst with the CBox tall block mesh */
  ntris = cbox_block_ntris;
  nverts = cbox_block_nverts;
  desc.vertices = cbox_short_block;
  desc.indices = cbox_block_ids;
  CHECK(s3d_shape_create_mesh(dev, &tall_block), RES_OK);
  CHECK(s3d_shape_get_id(tall_block, &tall_block_id), RES_OK);
  CHECK(s3d_mesh_setup_indexed_vertices
    (tall_block, ntris, cbox_get_ids, nverts, &attribs, 1, &desc), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, tall_block), RES_OK);

  /* Update the inst vertices */
  desc.vertices = cbox_tall_block;
  CHECK(s3d_mesh_setup_indexed_vertices
    (tall_block, ntris, S3D_KEEP, nverts, &attribs, 1, &desc), RES_OK);

  /* Create a the CBox short block inst */
  desc.vertices = cbox_short_block;
  CHECK(s3d_shape_create_mesh(dev, &short_block), RES_OK);
  CHECK(s3d_shape_get_id(short_block, &short_block_id), RES_OK);
  CHECK(s3d_mesh_setup_indexed_vertices
    (short_block, ntris, cbox_get_ids, nverts, &attribs, 1, &desc), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, short_block), RES_OK);

  /* Instantiate the scene */
  CHECK(s3d_scene_instantiate(scn, &inst), RES_OK);
  CHECK(s3d_scene_begin_session(scn, S3D_SAMPLE), RES_OK);
  CHECK(s3d_scene_primitives_count(scn, &nprims), RES_OK);
  CHECK(nprims, 20);
  CHECK(s3d_scene_end_session(scn), RES_OK);
  CHECK(s3d_shape_get_id(inst, &inst_id), RES_OK);

  /* Create the CBox walls */
  desc.indices = cbox_walls_ids;
  desc.vertices = cbox_walls;
  nverts = cbox_walls_nverts;
  ntris = cbox_walls_ntris;
  CHECK(s3d_shape_create_mesh(dev, &walls), RES_OK);
  CHECK(s3d_shape_get_id(walls, &walls_id), RES_OK);
  CHECK(s3d_mesh_setup_indexed_vertices
    (walls, ntris, cbox_get_ids, nverts, &attribs, 1, &desc), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, walls), RES_OK);

  /* Check that the ids are all different */
  NCHECK(walls_id, short_block_id);
  NCHECK(walls_id, tall_block_id);
  NCHECK(short_block_id, tall_block_id);

  /* Attach the CBox instance to a scene */
  CHECK(s3d_scene_create(dev, &scn2), RES_OK);
  f3(org, -100.f, 0.f, -2.f);
  CHECK(s3d_scene_attach_shape(scn2, inst), RES_OK);
  CHECK(s3d_instance_set_position(inst, org), RES_OK);

  CHECK(s3d_shape_enable(inst, 0), RES_OK);

  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_begin_session(scn2, S3D_TRACE), RES_BAD_OP);
  CHECK(s3d_scene_end_session(scn), RES_OK);
  CHECK(s3d_scene_begin_session(scn2, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_BAD_OP);
  CHECK(s3d_scene_end_session(scn), RES_BAD_OP);

  CHECK(s3d_scene_clear(scn2), RES_BAD_OP);
  CHECK(s3d_scene_clear(scn), RES_BAD_OP);

  CHECK(s3d_scene_end_session(scn2), RES_OK);
  CHECK(s3d_shape_enable(inst, 1), RES_OK);
  CHECK(s3d_scene_begin_session(scn2, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_end_session(scn2), RES_OK);

  CHECK(s3d_shape_create_mesh(dev, &walls_copy), RES_OK);
  CHECK(s3d_mesh_copy(walls, walls_copy), RES_OK);
  CHECK(s3d_shape_ref_put(walls), RES_OK);
  CHECK(s3d_shape_get_id(walls_copy, &walls_id), RES_OK);
  if(filter) {
    CHECK(s3d_mesh_set_hit_filter_function
      (walls_copy, filter_func, (void*)(uintptr_t)0xDECAFBAD), RES_OK);
  }

  CHECK(s3d_scene_clear(scn), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, walls_copy), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, short_block), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, tall_block), RES_OK);

  CHECK(s3d_scene_begin_session(scn2, S3D_TRACE|S3D_GET_PRIMITIVE), RES_OK);
  CHECK(s3d_scene_primitives_count(scn2, &nprims), RES_OK);
  CHECK(nprims, 30);

  CHECK(s3d_scene_get_aabb(scn2, lower, upper), RES_OK);
  CHECK(eq_epsf(lower[0], -100.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[1], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[2], -2.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[0], 452.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[1], 559.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[2], 546.f, 1.e-6f), 1);

  FOR_EACH(i, 0, nprims) {
    size_t j;
    CHECK(s3d_scene_get_primitive(scn2, (unsigned)i, prims + i), RES_OK);
    CHECK(S3D_PRIMITIVE_EQ(prims + i, &S3D_PRIMITIVE_NULL), 0);
    FOR_EACH(j, 0, i) {
      CHECK(S3D_PRIMITIVE_EQ(prims + i, prims + j), 0);
    }
  }

  camera_init(&cam);
  FOR_EACH(iy, 0, IMG_HEIGHT) {
    float pixel[2];

    pixel[1] = (float)iy/(float)IMG_HEIGHT;
    FOR_EACH(ix, 0, IMG_WIDTH) {
      const size_t ipix = (iy*IMG_WIDTH + ix) * 3/*RGB*/;

      pixel[0] = (float)ix/(float)IMG_WIDTH;
      camera_ray(&cam, pixel, org, dir);
      CHECK(s3d_scene_trace_ray
        (scn2, org, dir, range, (void*)(uintptr_t)0xDEADBEEF, &hit), RES_OK);

      if(S3D_HIT_NONE(&hit)) {
        if(img) {
          img[ipix+0] = img[ipix+1] = img[ipix+2] = 0;
        }
      } else {
        float N[3], len, dot, col[3] = { 1.f, 1.f, 1.f };
        struct s3d_attrib attr;
        float pos[3];

        CHECK(hit.prim.inst_id, inst_id);
        CHECK(hit.prim.geom_id == walls_id
           || hit.prim.geom_id == tall_block_id
           || hit.prim.geom_id == short_block_id, 1);
        CHECK(hit.prim.geom_id < 10, 1);

        CHECK(s3d_primitive_get_attrib
          (&hit.prim, S3D_POSITION, hit.uv, &attr), RES_OK);
        CHECK(attr.type, S3D_FLOAT3);
        CHECK(attr.usage, S3D_POSITION);
        f3_add(pos, f3_mulf(pos, dir, hit.distance), org);
        CHECK(f3_eq_eps
          (pos, attr.value, 1.e-3f/*Sic O_o!! Really bad precision!*/), 1);

        len = f3_normalize(N, hit.normal);
        NCHECK(len, 0);

        CHECK(s3d_primitive_get_attrib
          (&hit.prim, S3D_GEOMETRY_NORMAL, hit.uv, &attr), RES_OK);
        CHECK(attr.type, S3D_FLOAT3);
        CHECK(attr.usage, S3D_GEOMETRY_NORMAL);
        f3_normalize(attr.value, attr.value);
        CHECK(f3_eq_eps(attr.value, N, 1.e-6f), 1);

        CHECK(hit.prim.scene_prim_id >= hit.prim.prim_id, 1);
        CHECK(hit.prim.scene_prim_id < 30, 1);

        if(!img)
          continue;

        if(hit.prim.geom_id == walls_id) {
          if(hit.prim.prim_id == 4 || hit.prim.prim_id == 5) {
            col[0] = 1.f, col[1] = 0.f, col[2] = 0.f;
          } else if(hit.prim.prim_id == 6 || hit.prim.prim_id == 7) {
            col[0] = 0.f, col[1] = 1.f, col[2] = 0.f;
          }
        }

        dot = f3_dot(N, dir);
        if(dot < 0.f)
          dot = f3_dot(f3_minus(N, N), dir);

        img[ipix+0] = (unsigned char)(dot * col[0] * 255.f);
        img[ipix+1] = (unsigned char)(dot * col[1] * 255.f);
        img[ipix+2] = (unsigned char)(dot * col[2] * 255.f);
      }
    }
  }
  CHECK(s3d_scene_end_session(scn2), RES_OK);

  if(img_name) {
    CHECK(image_ppm_write(img_name, IMG_WIDTH, IMG_HEIGHT, 3, img), RES_OK);
  }

  if(img)
    MEM_RM(&allocator, img);

  CHECK(s3d_device_ref_put(dev), RES_OK);
  CHECK(s3d_shape_ref_put(inst), RES_OK);
  CHECK(s3d_shape_ref_put(short_block), RES_OK);
  CHECK(s3d_shape_ref_put(tall_block), RES_OK);
  CHECK(s3d_shape_ref_put(walls_copy), RES_OK);
  CHECK(s3d_scene_ref_put(scn), RES_OK);
  CHECK(s3d_scene_ref_put(scn2), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);

  return 0;
}

